package org.netkernel.layer0.meta.impl;

import java.util.List;

import javax.xml.parsers.ParserConfigurationException;

import org.netkernel.container.IMessages;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.GrammarParseException;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.urii.INetKernelException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class MetadataUtils
{
	private static final PairList VALIDATE_ACTIVE_GRAMMAR = new PairList(5);
	static
	{	VALIDATE_ACTIVE_GRAMMAR.put("*[name()!='identifier' and name()!='argument' and name()!='varargs']", "MSG_GRAMMAR_ACTIVE_ROOT_TAGS");
		VALIDATE_ACTIVE_GRAMMAR.put("../active[count(identifier)=0]", "MSG_GRAMMAR_ACTIVE_NO_IDENTIFIER");
		VALIDATE_ACTIVE_GRAMMAR.put("*/*", "MSG_GRAMMAR_ACTIVE_NO_DEEPER_TAGS");
		VALIDATE_ACTIVE_GRAMMAR.put("argument/attribute::*[name()!='name' and name()!='desc' and name()!='max' and name()!='min']", "MSG_GRAMMAR_ACTIVE_ARG_ATTS");
		VALIDATE_ACTIVE_GRAMMAR.put("type/attribute::*", "MSG_GRAMMAR_ACTIVE_NO_DEEPER_ATTRIBUTES");
		VALIDATE_ACTIVE_GRAMMAR.put("varargs/attribute::*", "MSG_GRAMMAR_ACTIVE_NO_DEEPER_ATTRIBUTES");	
	};
	
	public static IEndpointArgumentMeta getEndpointIdentifierPart(IEndpointMeta aMeta, String aName)
	{	IEndpointArgumentMeta result=null;
		IEndpointArgumentMeta[] parts=aMeta.getArguments();
		for (int i=0; i<parts.length; i++)
		{	if (parts[i].getName().equals(aName))
			{	result=parts[i];
				break;
			}
		}
		return result;
	}
	
	public static boolean match(IEndpointMeta aEndpointMeta, INKFRequestReadOnly aRequest)
	{	boolean result=false;
		int verb=aRequest.getVerb();
		if ((verb&aEndpointMeta.getSupportedVerbs())!=0) // verb matches
		{	if (verb==INKFRequestReadOnly.VERB_TRANSREPRESENT)
			{	//inbound and outbound representations tested
				Class[] responseClasses=aEndpointMeta.getResponseClasses();
				Class requestedClass = aRequest.getRepresentationClass();
				int i;
				for (i=0; i<responseClasses.length; i++)
				{	if (requestedClass.isAssignableFrom(responseClasses[i]))
					{	break;
					}
				}
				if (responseClasses.length==0 || i<responseClasses.length)
				{	Object primary=aRequest.getPrimary();
					if (primary!=null)
					{	Class from = primary.getClass();
						IEndpointArgumentMeta parts[] = aEndpointMeta.getArguments();
						for (int j=0; j<parts.length; j++)
						{	Class fromClasses[] = parts[j].getRequiredRepresentations();
							for (i=0; i<fromClasses.length; i++)
							{	if (fromClasses[i].isAssignableFrom(from))
								{	result = true;
								}
							}
						}
					}
				}
			}
			else
			{	// other requests just look at identifier
				result = aEndpointMeta.getIdentifierGrammar().matches(aRequest.getIdentifier());
			}
		}
	
		return result;
	}
	
	public static void parseGrammar(Element aGrammarDefn, EndpointMetaBuilder aBuilder, IMessages aMessages) throws ParserConfigurationException, GrammarParseException, INetKernelException
	{
		Element activeGrammar = XMLUtils.getChildElementNamed(aGrammarDefn, "active");
		if (activeGrammar!=null)
		{	
			PairList validation=FastSchematron.validate(activeGrammar, VALIDATE_ACTIVE_GRAMMAR);
			if (validation.size()>0)
			{	String m = FastSchematron.formatValidationError(validation,aMessages);			
				throw Utils.createFormattedException("EX_STD_MAPPER_CONFIG", "MSG_RAW", aMessages, m);
			}
			
			XMLReadable r = new XMLReadable(activeGrammar);
		
			IGrammarGroup grammarRoot = BNFGrammarFactory.createGroup(null);
			List<String> identifiers=r.getTexts("identifier", true);
			int is=identifiers.size();
			
			//parse identifiers
			String[] schemes = new String[is];
			String[] types = new String[is];
			for (int i=0; i<is; i++)
			{	String identifier=identifiers.get(i);
				int j=identifier.indexOf(':');
				if (j>=0)
				{	schemes[i]=identifier.substring(0,j);
					types[i]=identifier.substring(j+1);
				}
				else
				{	schemes[i]="";
					types[i]=identifier;
				}
			}
			//check for scheme homogeneity
			boolean schemeHomogeneity=true;
			String firstScheme=schemes[0];
			for (int i=1; i<is; i++)
			{	if (!schemes[i].equals(firstScheme))
				{	schemeHomogeneity=false;
					break;
				}
			}
			
			//build appropriate structure
			if (schemeHomogeneity)
			{	IGrammarGroup scheme = BNFGrammarFactory.createGroup(grammarRoot,ParsedIdentifierImpl.ARG_ACTIVE_SCHEME,null);
				BNFGrammarFactory.createText(scheme, firstScheme);
				BNFGrammarFactory.createText(grammarRoot, ":");
				if (is>1)
				{	IGrammarGroup typeGroup = BNFGrammarFactory.createChoiceGroup(grammarRoot, ParsedIdentifierImpl.ARG_ACTIVE_TYPE,null);
					for (String typeValue: types)
					{	BNFGrammarFactory.createText(typeGroup, typeValue);
					}
				}
				else
				{	IGrammarGroup typeGroup=BNFGrammarFactory.createGroup(grammarRoot, ParsedIdentifierImpl.ARG_ACTIVE_TYPE,null);
					BNFGrammarFactory.createText(typeGroup, types[0]);
				}
			}
			else
			{	IGrammarGroup choices = BNFGrammarFactory.createChoiceGroup(grammarRoot);
				for (int i=0; i<is; i++)
				{	IGrammarGroup choice = BNFGrammarFactory.createGroup(choices);
					IGrammarGroup scheme = BNFGrammarFactory.createGroup(choice,ParsedIdentifierImpl.ARG_ACTIVE_SCHEME,null);
					BNFGrammarFactory.createText(scheme, schemes[i]);
					IGrammarGroup type = BNFGrammarFactory.createGroup(choice,ParsedIdentifierImpl.ARG_ACTIVE_SCHEME,null);
					BNFGrammarFactory.createText(type, types[i]);
				}
			}
			
			List arguments = r.getNodes("argument");
			boolean hasVarargs = r.getNodes("varargs").size()>0;
			IGrammarGroup interleave=null;
			if (arguments.size()>0 || hasVarargs)
			{	interleave=BNFGrammarFactory.createInterleave(grammarRoot);
			}
			
			for (int i=0; i<arguments.size(); i++)
			{	XMLReadable arg=new XMLReadable((Node)arguments.get(i));
				int min=1;
				int max=1;
				String t=arg.getText("@min");
				if (t.length()>0)
				{	min=Integer.parseInt(t);
				}
				t=arg.getText("@max");
				if (t.length()>0)
				{	if (t.equals("*"))
					{	max = Integer.MAX_VALUE;
					}
					else
					{	max=Integer.parseInt(t);
					}
				}
				String name=arg.getText("@name");
				String desc=arg.getText("@desc");
				if (desc.length()==0) desc=null;
				
				ActiveEndpointMetaBuilder.createArgument(interleave, name, min, max);		
				IEndpointArgumentMeta argMeta = new EndpointArgumentMetaImpl(name,desc,null,null,null,true,null);
				aBuilder.addArgument(argMeta);			
			}
			
			if (hasVarargs)
			{	ActiveEndpointMetaBuilder.createVarargs(interleave);				
			}
			aBuilder.setGrammarRoot(grammarRoot);		
		}
		else
		{	// low level grammar definition
			Document d=XMLUtils.newDocument();
			Element de = d.createElement("group");
			d.appendChild(de);
			Element e2=(Element)d.importNode(aGrammarDefn, true);
			Node n=e2.getFirstChild();
			while ( n!=null)
			{	de.appendChild(n);
				n=e2.getFirstChild();
			}
			IIdentifierGrammar grammar= BNFGrammarFactory.parse(de,aMessages);
			aBuilder.setGrammar(grammar);
		}		
	}
}
